#! /usr/bin/env python
# -*- coding: utf-8; -*-

'''
Neural field model of motion integration.
'''

# Copyright © INRIA Sophia Antipolis – Méditerranée
# Author: Émilien Tlapale <emilien@tlapale.com>
# 
# This software is a computer program whose purpose is to [describe
# functionalities and technical features of your software].
# 
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-C license and that you accept its terms.


from optparse import OptionParser
from os import environ
import os.path
from os.path import dirname
from signal import SIGINT, signal
import sys
from sys import exit
from time import clock, time

from numpy import *
from numpy.random import normal
from scipy import random
from scipy.integrate import ode
from scipy.ndimage.filters import *
from tables import openFile, Float32Atom


def sigmoid(x, steepness):
    return 1./(1. + exp(-steepness*(x)))# - 0.001

if __name__ == '__main__':
    # Parse options
    parser = OptionParser()
    parser.add_option('', '--tres', dest='tres', default=0.1,
                      help='Temporal resolution')
    parser.add_option('', '--l1l', dest='l1l', default=2.0,
                      help='Leak weight in p1')
    parser.add_option('', '--l1f', dest='l1f', default=1.0,
                      help='Feedforward weight in p1')
    parser.add_option('', '--l1b', dest='l1b', default=24.0,
                      help='Feedback weight in p1')
    parser.add_option('', '--l1i', dest='l1i', default=196.0,
                      help='Lateral inhibition weight in p1')
    parser.add_option('', '--l2l', dest='l2l', default=2.0,
                      help='Leak weight in p2')
    parser.add_option('', '--l2f', dest='l2f', default=16.0,
                      help='Feedforward weight in p2')
    parser.add_option('', '--l2i', dest='l2i', default=196.0,
                      help='Lateral inhibition weight in p2')
    parser.add_option('', '--s1i', dest='s1i', default=2.0,
                      help='Lateral inhibition extent in p1')
    parser.add_option('', '--s2i', dest='s2i', default=2.0,
                      help='Lateral inhibition extent in p2')
    parser.add_option('', '--s2f', dest='s2f', default=8.0,
                      help='Feedforward diffusion in p2')
    parser.add_option('-e', '--end', dest='maxdepth', default=500,
                      help='Maximum depth')
    parser.add_option('-s', '--steepness', dest='steepness', default=8.0,
                      help='Sigmoid steepness')
    parser.add_option('-a', '--p0-amp', dest='p0_amp', default='1.0',
                      help='Amplitude of p0 [1.0]')
    options, args = parser.parse_args()

    # Parameters
    tres = float(options.tres)
    c1, ca, cb, cc = [float(x) for x in \
                      [options.l1l, options.l1f, options.l1b, options.l1i]]
    c2, cm, cn = [float(x) for x in \
                  [options.l2l, options.l2f, options.l2i]]
    sg1, sg2 = [float(x) for x in [options.s1i, options.s2i]]
    p0_amp = float(options.p0_amp)
    sgx = float(options.s2f)
    max_depth = int(options.maxdepth)
    steepness = float(options.steepness)
    
    # Open the input data in R/W mode
    hf = openFile(args[0], 'a')
    h0 = hf.root.p0
    vdim, height, width = h0.shape[1:]
    depth = min(max_depth, h0.shape[0])
    
    # Parse parameters
    long_depth = int(depth / tres) + 1
    factor = tres # / tau
    
    # Initial conditions
    p12 = zeros((2, vdim, height, width), dtype=float32)

    # Output datasets
    o1 = hf.createCArray(hf.root, 'p1', Float32Atom(), [long_depth] + list(p12.shape[1:]))
    o2 = hf.createCArray(hf.root, 'p2', Float32Atom(), o1.shape)

    # Store parameters
    for k, v in [('c1', c1), ('ca', ca), ('cb', cb), ('cc', cc),
                 ('c2', c2), ('cm', cm), ('cn', cn),
                 ('sgx', sgx), ('sg1', sg1), ('sg2', sg2),
                 ('tres', tres), ('factor', factor),]:
        setattr(o1._v_attrs, k, v)
        setattr(o2._v_attrs, k, v)
    
    # Iterations loop
    def rhs(t, p, arg):
        print '.'
        # Fetch the arrays
        p_ = p.reshape(p12.shape)
        p1 = p_[0]
        p2 = p_[1]
    
        # p1 feedforward and feedback
        p1c = p0 * (ca + cb * p2)
    
        # p1 lateral inhibition
        p1i = gaussian_filter(average(p1, 0), sg1)
        for v in xrange(vdim):
            p1c[v] -= cc * p1i
        p1_incr = - c1 * p1 + sigmoid(p1c, steepness)
    
        # p2 feed-forward
        p2c = zeros(p1.shape, dtype=float32)
        for v in xrange(vdim):
            res = gaussian_filter1d(gaussian_filter1d(p1[v], sgx, 0), sgx, 1)
            p2c[v] = cm * res
    
        # p2 lateral inhibition
        p2i = gaussian_filter(average(p2, 0), sg2)
        for v in xrange(vdim):
            p2c[v] -= cn * p2i
        # Compute p2 increment
        p2_incr = - c2 * p2 + sigmoid(p2c, steepness)
    
        # Return
        ret = zeros_like(p12)
        ret[0] = p1_incr
        ret[1] = p2_incr
        return ret.flat

    # Input
    coarse_t = 0
    p0 = p0_amp * h0.read(coarse_t)[0]

    # Create the solver
    t0, t1 = 0.0, max_depth
    solver = ode(rhs).set_integrator('vode', method='adams',
                                     with_jacobian=False)
    solver.set_initial_value(p12.flat, t0).set_f_params(p0)

    # Integrate the equation
    oi=0
    while solver.successful() and solver.t < t1:
        print 'Time', solver.t
        # Input
        if coarse_t != int(solver.t):
            print '  Loading frame', int(solver.t)
            coarse_t = int(solver.t)
            p0 = p0_amp * h0.read(coarse_t)[0]

        # Solver
        solver.integrate(solver.t + tres)
        ans = solver.y.reshape(p12.shape)
    
        # Store output
        print 'Time %f storing %d' % (solver.t, oi)
        o1[oi] = ans[0]
        o2[oi] = ans[1]
        oi+=1
        hf.flush()

    hf.close()
