# -*- coding: utf-8; -*-

'''
Some generic utilities.
'''

# Copyright © INRIA Sophia Antipolis – Méditerranée
# Author: Émilien Tlapale <emilien@tlapale.com>
# 
# This software is a computer program whose purpose is to [describe
# functionalities and technical features of your software].
# 
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-C license and that you accept its terms.


from math import atan2, exp, pi, sqrt
import sys

from numpy import *
from tables import Float32Atom


def norm_angle(theta):
    'Normalize an angle between 0 and 2π.'
    if theta < 0.0:
        return fmod(theta, 2.0 * pi) + 2.0 * pi
    else:
        return fmod(theta, 2.0 * pi)


def calculate_of(data, vspace, threshold, epsilon=1e-20):
    '''
    Calculate the optical flow of a multiple velocity field.
    '''
    # Put v as last dim for dot product
    p = swapaxes(swapaxes(data, 0, 1), 1, 2)
    # Calculate the dot product
    of_x = dot(p, vspace.real) #vspace[...,1])
    of_y = dot(p, vspace.imag) #vspace[...,0])
    # Calculate the activity sums
    vsums = sum(p, 2) + epsilon
    # Normalize the results
    of_x /= vsums
    of_y /= vsums
     # Threshold according to activity sum
    of_x *= (vsums > threshold)
    of_y *= (vsums > threshold)

    return (of_x, of_y)


def parse_vspace(arg, sep1=None, sep2=',', sub=1):
    '''
    Parse a string argument to a velocity space array.

    The string argument may either be a single number
    to build a linear cartesian velocity space with this
    value for maximum X- and Y-coordinate offset,
    or a whitespace (sep1) separated list of coma (sep2) separated
    floating point numbers encoding coordinates.
    The X-coordinate offset is always specified before
    the Y-coordinate one, in a (v_x,v_y) fashion.

    Return an unsorted complex 1D array containing
    the velocities. Real part contains the X-coordinate
    offset while Y-coordinate offset is imaginary.

    :param arg: String to be converted to vspace.
    :param sep1: String separating velocities.
    :param sep2: String separating veclocity components.
    :param sub: Subpixel sampling.
    :return: A 1D complex ndarray containg velocities.
    '''
    if isinstance(arg, int) or (isinstance(arg, str) and arg.isdigit()):
        n = int(arg)
        b, a = ogrid[-n:n+sub:sub,-n:n+sub:sub]
        return (a+1j*b).reshape(int(((n/sub)*2+1)**2))
    else:
        return array([complex(*[floatter(x) for x in a.split(sep2)])
                      for a in arg.split(sep1)])


def draw_color_disk(cr, x, y, width, height, cmap):
    from cairo import ImageSurface, FORMAT_ARGB32

    width, height=400, 400
    cx, cy = int(width/2), int(height/2)
    max_rho = min(cx, cy) # sqrt((w/2)**2+(h/2)**2)

    # Create the cairo surface and associated NumPy array
    sur = ImageSurface(FORMAT_ARGB32, int(width), int(height))
    a = frombuffer(sur.get_data(), uint8)
    a.shape = (height, width, 4)

    # Draw the color disk
    for j in xrange(int(height)):
        ry = j - cy
        for i in xrange(int(width)):
            rx = i - cx
            theta = norm_angle(atan2(ry, rx))
            rho = sqrt(rx**2 + ry**2)
            if rho <= max_rho:
              r, g, b = cmap(theta/(2*pi), rho/max_rho)
              a[j, i] = [b, g, r, 255]
            else:
              a[j, i] = [255, 255, 255, 0]
    sur.mark_dirty()
    #sur.write_to_png('circle.png')

    # Clip the color disk with anti-aliasing
    #cr.arc(x, y, max_rho - 1, 0, 2*pi)
    cr.set_source_surface(sur)
    #cr.arc(x, y, max_rho - 1, 0, 2*pi)
    cr.paint()
