# -*- mode: python; coding: utf-8; -*-

'''
Various kind of useful colormaps.

Colormaps dealing with (theta, rho) coordinates rely on their
arguments to be normalized in [0..1]. They return a 3-uple
containing (R, G, B) color components as integers in [0..255].
'''

# Copyright © INRIA Sophia Antipolis – Méditerranée
# Author: Émilien Tlapale <emilien@tlapale.com>
# 
# This software is a computer program whose purpose is to [describe
# functionalities and technical features of your software].
# 
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-C license and that you accept its terms.


from colorsys import hls_to_rgb
from numpy import array, ndarray


def middleburry(theta, rho):
    '''
    Middleburry colorcode used in [Baker-etal-2007]_ for optical flows.
    It is based on Munsell and J. Savard work on color perception.
    
    .. [Baker-etal-2007] S. Baker, D. Scharstein,
       J.P. Lewis, S. Roth, M.J. Black and R. Szeliski,
       *A Database and Evaluation Methodology for Optical
       Flow*, ICCV, 2007.
    '''
    # If necessary, initialize the color wheel
    if middleburry.c_wheel is None:
        ry = 15
        yg = 6
        gc = 4
        cb = 11
        bm = 13
        mr = 6
        ncols = ry + yg + gc + cb + bm + mr
        
        middleburry.c_wheel = ndarray((3, ncols), dtype='float32')
        
        k = 0
        for i in range(ry):
            middleburry.c_wheel[:,k] = 255, 255*i/ry, 0
            k += 1
        for i in range(yg):
            middleburry.c_wheel[:,k] = 255-255*i/yg, 255, 0
            k += 1
        for i in range(gc):
            middleburry.c_wheel[:,k] = 0, 255, 255*i/gc
            k += 1
        for i in range(cb):
            middleburry.c_wheel[:,k] = 0, 255-255*i/cb, 255
            k += 1
        for i in range(bm):
            middleburry.c_wheel[:,k] = 255*i/bm, 0, 255
            k += 1
        for i in range(mr):
            middleburry.c_wheel[:,k] = 255, 0, 255-255*i/mr
            k += 1
    # Use the existing color wheel
    fk = theta * (middleburry.c_wheel.shape[1]-1)
    k0 = fk.astype('int')#int(fk)
    k1 = (k0 + 1) % middleburry.c_wheel.shape[1]
    f = fk - k0
    
    col0 = middleburry.c_wheel[:,k0] / 255.
    col1 = middleburry.c_wheel[:,k1] / 255.
    print f.shape, col0.shape, col1.shape
    col = (1. - f) * col0 + f * col1
    # Increase saturation with radius
    col = 1. - rho * (1. - col)
    return array(255. * col, dtype='uint8')

# Mark middleburry color wheel as uninitialized    
middleburry.c_wheel = None


def hls(theta, rho):
    '''
    HLS based colorcode used in [Tlapale-etal-2007]_.

    .. [Tlapale-etal-2007] É. Tlapale, G.S. Masson and
       P. Kornprobst, *Motion integration modulated by form
       information*, Neurocomp, 2008.
    '''
    return [int(255*x) for x in hls_to_rgb(theta, .5, rho)]
