#! /usr/bin/env python
# -*- coding: utf-8; -*-

"""
Bayerl & Neumann's modified Reichardt motion detectors.
"""

# Copyright © INRIA Sophia Antipolis – Méditerranée
# Author: Émilien Tlapale <emilien@tlapale.com>
# 
# This software is a computer program whose purpose is to [describe
# functionalities and technical features of your software].
# 
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-C license and that you accept its terms.


from math import degrees
from os import environ
from optparse import OptionParser

from numpy import *
from scipy.ndimage.filters import gaussian_filter1d, gaussian_filter

from tables import openFile, Float32Atom

from pymotion import parse_vspace


if __name__ == '__main__':
    # Parse the command line arguments
    parser = OptionParser()
    parser.add_option('-V', '--vspace', dest='vspace', default='3',
                      help='Velocity space to use')
    options, args = parser.parse_args()

    # Parameters
    vspace = parse_vspace(options.vspace)
    ddim = 8
    sgd, sgb, sgh = 1.0, 4.0, 2.0
    epsilon = 1e-20

    # Open the input data in R/W mode
    hf = openFile(args[0], 'a')
    hs = hf.root.stimulus
    depth, height, width = hs.shape

    # Check the vspace
    vdim = len(vspace)
    vx_min = vx_max = vy_min = vy_max = 0
    for (vx, vy) in [(x.real, x.imag) for x in vspace]:
        vy_min = min(vy_min, vy)
        vx_min = min(vx_min, vx)
        vy_max = max(vy_max, vy)
        vx_max = max(vx_max, vx)
    ext_height, ext_width = vy_max + height - vy_min, vx_max + width - vx_min

    # Results
    dderiv = zeros((ddim, height, width))
    c1 = zeros((ddim, height, width))
    op0 = hf.createCArray(hf.root, 'p0', Float32Atom(), [depth - 1, vdim, height, width])

    # Store parameters
    op0._v_attrs.sigma_deriv = sgd
    op0._v_attrs.sigma_normalizing = sgb
    op0._v_attrs.sigma_detectors = sgh

    # Iterations loop
    ext_c1 = None
    for t in range(depth):
        print '[%d/%d]' % (t + 1, depth)
        
        frame = hs.read(t)[0]
        #prev_dderiv = dderiv.copy()

        # Compute the derivatives on X and Y
        dx = gaussian_filter1d(frame, sgd, 1, 1)
        dy = gaussian_filter1d(frame, sgd, 0, 1)
        dxx = gaussian_filter1d(dx, sgd, 1, 1)
        dxy = gaussian_filter1d(dx, sgd, 0, 1)
        dyy = gaussian_filter1d(dy, sgd, 0, 1)
        
        # Compute directional derivatives
        for d in range(ddim):
            theta = d * pi / ddim
            cos_theta = cos(theta)
            sin_theta = sin(theta)
            cos_cos = cos_theta * cos_theta
            sin_sin = sin_theta * sin_theta
            cos_sin = cos_theta * sin_theta
            
            dderiv[d] = cos_cos * dxx \
                + 2 * cos_sin * dxy \
                + sin_sin * dyy
        
        # Normalize derivatives through division
        ddsum = gaussian_filter(abs(dderiv).sum(0), sgb)
        for d in range(ddim):
            c1[d] = dderiv[d] / (epsilon + ddsum)

        # Zeros surround for dderiv
        ext_prev_c1 = ext_c1
        ext_c1 = zeros((ddim, ext_height, ext_width))
        for d in xrange(ddim):
            ext_c1[d, vy_max:vy_max+height, vx_max:vx_max+width] = c1[d]

        # We have to skip a frame
        if t == 0:
            continue
        
        # Half detectors
        hplus = zeros((vdim, height, width))
        hminus = zeros((vdim, height, width))    
        for v in xrange(vdim):
            vx, vy = vspace[v].real, vspace[v].imag
            plus = zeros((height, width))
            minus = zeros((height, width))
            for d in xrange(ddim):
                plus += ext_prev_c1[d,vy_max:vy_max+height,vx_max:vx_max+width] \
                        * ext_c1[d,vy_max+vy:vy_max+vy+height,vx_max+vx:vx_max+vx+width]
                minus += c1[d] * ext_prev_c1[d,vy_max+vy:vy_max+vy+height,vx_max+vx:vx_max+vx+width]
            
            hplus[v] = gaussian_filter(plus, sgh)
            hminus[v] = gaussian_filter(minus, sgh)
        
        # Combine the half detectors
        hminus *= (hminus>0)
        p0 = (hplus*(hplus>0) - 0.5 * hminus) / (1.0 + hminus)
        p0 *= (p0>0)
        
        # Save to disk
        op0[t - 1] = p0

    hf.close()
