#! /usr/bin/env python
# -*- coding: utf-8; -*-

'''
Flow displayer utility.
'''

# Copyright © INRIA Sophia Antipolis – Méditerranée
# Author: Émilien Tlapale <emilien@tlapale.com>
# 
# This software is a computer program whose purpose is to [describe
# functionalities and technical features of your software].
# 
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-C license and that you accept its terms.


from math import atan2, degrees, pi
from optparse import OptionParser
from sys import stderr
from time import clock

import cairo
from cairo import Context, FORMAT_ARGB32, HAS_PDF_SURFACE, ImageSurface
import gtk

from numpy import abs, arange, array, cos, dot, frombuffer, sin, sqrt, sum, swapaxes, uint8
from tables import openFile

from pymotion import calculate_of, norm_angle, parse_vspace, draw_color_disk
from pymotion.colormap import middleburry, hls

def draw_generic_arrow(cr, x, y, theta, rho,
                       center_size, head_size, line_width,
                       head_ratio=4.0):
    '''
    Draw an arrow on the given Cairo context.
    '''
    cr.save()
    
    cr.set_source_rgba(0.0, 0.0, 0.0, 1.0)
    # Center
    if center_size > 0:
        cr.arc(x, y, 2.0, 0, 2*pi)
        cr.fill()
    # Make a (x,y) center
    #cr.translate(-(x + rho), -y)
    # Line
    cr.set_line_width(line_width)
    cr.move_to(x, y)
    cr.rotate(theta)
    cr.rel_line_to(rho, 0)
    cr.stroke()
    cr.rotate(-theta)
    # Head
    cr.move_to(x, y)
    cr.rotate(theta)
    cr.rel_move_to(rho, 0)
    cr.rel_move_to(head_size, 0)
    cr.rel_line_to(-head_ratio*head_size, head_size)
    cr.rel_line_to(0, -2*head_size)
    cr.rel_line_to(head_ratio*head_size, head_size)
    cr.fill()
    
    cr.restore()


def draw_arrow_1(cr, x, y, theta, rho):
    draw_generic_arrow(cr, x, y, theta, rho,
                       2.0, 2.0, 1.0)

def draw_arrow_2(cr, x, y, theta, rho):
    draw_generic_arrow(cr, x, y, theta, rho,
                       0.0, 4.0, 1.0)

def draw_arrow_3(cr, x, y, theta, rho):
    draw_generic_arrow(cr, x, y, theta, rho,
                       0.0, 4.0, 1.0, head_ratio=2.0)

def draw_arrow_4(cr, x, y, theta, rho):
    draw_generic_arrow(cr, x, y, theta, rho,
                       0.0, 4.0, 1.0, head_ratio=1.0)

def draw_arrow_5(cr, x, y, theta, rho):
    draw_generic_arrow(cr, x, y, theta, rho,
                       0.0, 6.0, 3.0)


def draw_triangle(cr, x, y, theta, rho):
    '''
    Draw a triangle whose size represent velocity magnitude.
    '''
    cr.save()
    cr.set_source_rgba(0.0, 0.0, 0.0, 1.0)
    a, b = rho * cos(theta), rho * sin(theta)
    # Triangle
    cr.move_to(x, y)
    cr.rotate(theta)
    cr.rel_line_to(-4*rho, rho)
    cr.rel_line_to(0, -2*rho)
    cr.rel_line_to(4*rho, rho)
    cr.fill()
    cr.restore()

arrow_styles = {
    'triangles': draw_triangle,
    'arrow_1': draw_arrow_1,
    'arrow_2': draw_arrow_2,
    'arrow_2': draw_arrow_2,
    'arrow_3': draw_arrow_3,
    'arrow_4': draw_arrow_4,
    'arrow_5': draw_arrow_5,
}


def colorize_of(sur, of, rho, vthreshold, cmap):
    '''
    Colorize an OF inside a Cairo image surface.
    '''
    # Get a NumPy array from Cairo surface
    cr = Context(sur)
    a = frombuffer(sur.get_data(), uint8)
    width, height = sur.get_width(), sur.get_height()
    a.shape = (height, width, 4)
    
    of_x, of_y = of
    #for y in xrange(height):
    #    for x in xrange(width):
    #        vx, vy = of_x[y, x], of_y[y, x]
    #        # Set the pixel color according to vx,vy
    #        if abs(vx) < vthreshold and abs(vy) < vthreshold:
    #            a[y, x] = 255, 255, 255, 255
    #        else:
    #            theta = norm_angle(atan2(vy, vx))
    #            r, g, b = cmap(theta/(2*pi), rho[y,x])
    #            a[y, x] = [b, g, r, 255]
    theta = arctan2(of_y, of_x)
    # Normalise the angles
    theta += (theta < 0) * (2*pi)
    colored = cmap(theta/(2*pi), rho)
    a[...,0] = colored[2]
    a[...,1] = colored[1]
    a[...,2] = colored[0]
    a[...,3] = 255

def paint_flow(cr, data, of, vspace,
               multi, of_bg, color_disk,
               scaling, sampling, arrow_length,
               threshold, vthreshold,
               config):
    '''
    Paint the given velocity field in the given Cairo context.
    '''
    vdim, height_in, width_in = data.shape
    width_out, height_out = [scaling * x for x in [width_in, height_in]]
    
    # Paint colored OF background
    print '  Background...'
    beg = clock()
    if of_bg and (of is not None):
        # Compute OF norm
        rho = sqrt(of[0]**2 + of[1]**2)
        rho /= rho.max() + 1e-20
        # Colorize the OF
        sur = ImageSurface(FORMAT_ARGB32, width_in, height_in)
        colorize_of(sur, of, rho, float(options.velocity_threshold), options.cmap)
        sur.mark_dirty()
        # Paint it as background
        factor = width_out / width_in
        cr.scale(factor, factor)
        cr.set_source_surface(sur, 0, 0)
        cr.paint()
        cr.scale(1./factor, 1./factor)
    # Fill background
    else:
        cr.rectangle(0, 0, width_out, height_out)
        cr.set_source_rgb(1, 1, 1)
        cr.fill()
    end = clock()
    print '    bg: %f' % (end - beg)

    # Draw the color disk
    if color_disk:
        print '  Color disk...'
        beg = clock()
        cw, ch = width_out/8, height_out/8
        x, y = cw/2, height_out - ch/2
        draw_color_disk(cr, x, y, cw, ch, options.cmap)
        end = clock()
        print '    cd: %f' % (end - beg)
    
    # Draw the arrows
    beg = clock()
    arrow_style = config['arrow_style']
    if arrow_style is None:
        return
    print '  Arrows...', height_in, sampling
    for y in xrange(0, int(height_in), int(sampling)):
        for x in xrange(0, int(width_in), int(sampling)):
            x_o, y_o = scaling * x, scaling * y
            # Draw multiple arrows
            if multi and not options.of_mode:
                for v in xrange(vdim):
                    if data[v, y, x] > threshold:
                        print '>>>>'
                        vx, vy =vspace[v].real, vspace[v].imag
                        theta = norm_angle(atan2(vy, vx))
                        rho = arrow_length * sqrt(vx**2+vy**2)
                        arrow_styles[arrow_style](cr, x_o, y_o, theta, rho)
                        #if triangles:
                        #if arrow_style == 'triangles':
                        #    draw_triangle(cr, x_o, y_o, theta, rho)
                        #elif arrow_style == 
                        #else:
                        #    draw_arrow(cr, x_o, y_o, theta, rho)
            # Draw single arrow for OF
            else:
                vx, vy = of[0][y, x], of[1][y, x]
                if abs(vx) > vthreshold or abs(vy) > vthreshold:
                    theta = norm_angle(atan2(vy, vx))
                    rho = arrow_length * sqrt(vx**2+vy**2)
                    arrow_styles[arrow_style](cr, x_o, y_o, theta, rho)
                    #if triangles:
                    #    draw_triangle(cr, x_o, y_o, theta, rho)
                    #else:
                    #    draw_arrow(cr, x_o, y_o, theta, rho)
    end = clock()
    print '    ar: %f' % (end - beg)

from pylab import *

class VSpaceViewer(gtk.DrawingArea):
    def __init__(self, vspace=None, data=None, interpolation='grid'):
        print 'Init'
        gtk.DrawingArea.__init__(self)
        print 'Ouch'
        self.interpolation = interpolation
        print 'daaloooofs'
        self.vspace = vspace
        print 'dqqqfs'
        self.set_data(data)
        print 'sdfjwedfs'
        self.connect('expose_event', self.expose_cb)
        print 'Initialized'

    def set_data(self, data):
        from scipy.interpolate import interp2d

        self.data = data

        if data is None:
            return

        # Interpolate values
        if self.interpolation == 'grid':
            self.interp = array(data)
            self.interp.shape = (sqrt(self.data.shape[0]), sqrt(self.data.shape[0]))
            #self.interp = self.interp.swapaxes(0, 1)
        else:
            interp = interp2d(self.vspace.real, self.vspace.imag, self.data,
                              kind=self.interpolation)
            xmax = max(abs(self.vspace.real))
            xoff = xmax / (2*len(self.vspace.real))
            x = arange(-xmax, xmax + xoff, xoff)
            self.interp = interp(x, x)

        # Normalize
        m, M = self.interp.min(), self.interp.max()
        self.interp = (self.interp - m) / (M - m)

        # Create Cairo surface
        self.sur = ImageSurface(FORMAT_ARGB32, self.interp.shape[0],
                                self.interp.shape[1])
        self.colorize_vspace()

        # Setup a minimal size
        self.set_size_request(*self.interp.shape)

        self.queue_draw()

    def set_interpolation(self, interpolation):
        self.interpolation = interpolation
        self.set_data(self.data)

    def expose_cb(self, area, evt):
        width_out, height_out = self.get_allocation().width, \
            self.get_allocation().height
        cr = self.window.cairo_create()

        if self.data is None:
            return
        
        anti = cr.get_antialias()
        cr.set_antialias(cairo.ANTIALIAS_NONE)

        # Paint it as background
        width_in, height_in = self.sur.get_width(), self.sur.get_height()
        #factor = min(width_out / width_in, height_out / height_in)
        factor = float(width_out) / width_in
        cr.scale(factor, factor)
        cr.set_source_surface(self.sur, 0, 0)
        cr.paint()
        cr.scale(1./factor, 1./factor)

        cr.set_antialias(anti)
    
    def colorize_vspace(self):
        width, height = self.sur.get_width(), self.sur.get_height()
        a = frombuffer(self.sur.get_data(), uint8)
        a.shape = (height, width, 4)
        a[...,0] = 255*self.interp
        a[...,1] = 255*self.interp
        a[...,2] = 255*self.interp
        a[...,3] = 255
        self.sur.mark_dirty()


def run_gui(dset, options, config):
    '''
    Run the visualization inside an interactive GUI.
    '''
    # Check for GTK+ 2.0
    from pygtk import require as pygtk_require
    pygtk_require('2.0')
    from gtk import Adjustment, ToolButton, CheckButton, DrawingArea, \
         FILE_CHOOSER_ACTION_SAVE, FileChooserDialog, FileFilter, \
         HandleBox, HBox, HScale, Label, \
         RESPONSE_ACCEPT, RESPONSE_REJECT, STOCK_CANCEL, STOCK_FLOPPY, \
         ScrolledWindow, SpinButton, VBox, Window, \
         combo_box_new_text, main, main_quit
    import gtk
    import gtk.gdk
         

    #depth, vdim, height_in, width_in = dset.shape

    # Init some GUI variables
    options.cwd_uri = None
    
    # Create main GUI window
    win = Window()
    win.connect('delete_event', main_quit)
    win.connect('destroy', main_quit)
    vbox = VBox()
    win.add(vbox)

    print 'Fill window'
    
    # Toolbar for parametrisation
    controls = HBox()
    hbox = HandleBox()
    hbox.add(controls)
    vbox.pack_start(hbox, False)
    time_adj = Adjustment(0, 0, dset.shape[0] - 1, 1)
    time_scale = HScale(time_adj)
    time_scale.set_digits(0)
    time_scale.set_value(int(options.begin))
    controls.pack_start(Label('Frame'), False)
    controls.pack_start(time_scale)
    def time_changed(scale):
        if options.frame_vspace:
            t = int(time_scale.get_value())
            f_vspace.set_data(dset[t].sum(1).sum(1))
        if config['painting']:
            update_buffer()
    time_scale.connect('value-changed', time_changed)
    # Save as button
    but = ToolButton(STOCK_FLOPPY)
    controls.pack_start(but, False)
    def save_frame(but):
        fc = FileChooserDialog('Save frame view', win, FILE_CHOOSER_ACTION_SAVE,
                               (STOCK_CANCEL, RESPONSE_REJECT,
                                STOCK_FLOPPY, RESPONSE_ACCEPT))
        if options.cwd_uri:
            fc.set_current_folder_uri(options.cwd_uri)
        ftr = FileFilter()
        ftr.set_name('PNG image')
        ftr.add_mime_type('image/png')
        ftr.add_pattern('*.png')
        fc.add_filter(ftr)
        if HAS_PDF_SURFACE:
            ftr = FileFilter()
            ftr.set_name('PDF document')
            ftr.add_mime_type('application/pdf')
            ftr.add_pattern('*.pdf')
            fc.add_filter(ftr)
        
        ans = fc.run()
        fc.hide()
        path = fc.get_filename()
        if ans == RESPONSE_ACCEPT and path:
            options.cwd_uri = fc.get_current_folder_uri()
            path = fc.get_filename()
            if path.endswith('.png') or path.endswith('.PNG'):
                options.sur.write_to_png(path)
            elif path.endswith('.pdf') or path.endswith('.PDF'):
                from cairo import PDFSurface
                width_out, height_out = [options.scaling * x \
                                         for x in [width_in, height_in]]
                sur = PDFSurface(path, width_out, height_out)
                cr = Context(sur)
                t = int(time_scale.get_value())
                print 'Frame', t
                paint_flow(cr, dset[t], options.of, options.vspace,
                           multi=options.multi, of_bg=options.of_bg,
                           color_disk=options.color_disk,
                           scaling=options.scaling, sampling=options.sampling,
                           arrow_length=options.arrow_length,
                           threshold=float(options.activity_threshold),
                           vthreshold=float(options.velocity_threshold),
                           config=config)
                cr.show_page()
                sur.finish()
            else:
                from sys import stderr
                print >>stderr, 'Unknown file type:', path
    but.connect('clicked', save_frame)
    # Scaling parameter
    controls = HBox()
    hbox = HandleBox()
    hbox.add(controls)
    vbox.pack_start(hbox, False)
    zoom_spin = SpinButton(Adjustment(options.scaling, 0.01, 1000.0, 1.0))
    zoom_spin.set_digits(1)
    controls.pack_start(Label('Scale'), False)
    controls.pack_start(zoom_spin, False)
    def zoom_changed(but):
        options.scaling = but.get_value()
        width_out, height_out = [int(options.scaling * x) \
                                 for x in [width_in, height_in]]
        area.set_size_request(width_out, height_out)
        if width_out != options.sur.get_width() \
               or height_out != options.sur.get_height():
            options.sur = sur = ImageSurface(FORMAT_ARGB32, width_out, height_out)
            options.cr = Context(sur)
        if config['painting']:
            update_buffer()
    zoom_spin.connect('value-changed', zoom_changed)
    # Generic parameters
    def add_spin(label, option, digits=0):
        spin = SpinButton(Adjustment(getattr(options, option), 0.001, 1000.0, 1.0))
        spin.set_digits(digits)
        controls.pack_start(Label(label), False)
        controls.pack_start(spin, False)
        def changed(but):
            setattr(options, option, but.get_value())
            if config['painting']:
                update_buffer()
        spin.connect('value-changed', changed)
    add_spin('A.len', 'arrow_length', 2)
    add_spin('Sampling', 'sampling')
    if not options.of_mode:
        add_spin('Threshold', 'activity_threshold', digits=4)
        add_spin('OF thr.', 'of_threshold', digits=4)
    add_spin('Vel. thr.', 'velocity_threshold', digits=4)
    # Generic options
    controls = HBox()
    hbox = HandleBox()
    hbox.add(controls)
    vbox.pack_start(hbox, False)
    def add_check(label, option):
        check = CheckButton(label)
        check.set_active(getattr(options, option))
        controls.pack_start(check, False)
        def changed(but):
            setattr(options, option, but.get_active())
            if config['painting']:
                update_buffer()
        check.connect('toggled', changed)
    # Arrow style list
    arrows_model = gtk.ListStore(gtk.gdk.Pixbuf, str)
    # Add the no arrows style
    arrows_model.append([None, 'none'])
    # Add all the arrow styles
    i=1
    active=0
    for k, v in arrow_styles.iteritems():
        w, h = 48, 24
        img = cairo.ImageSurface(cairo.FORMAT_ARGB32, w, h)
        cr = cairo.Context(img)
        v(cr, w/4, h/2, 0, w/2)
        pixbuf = gtk.gdk.pixbuf_new_from_data(
            img.get_data(),
            gtk.gdk.COLORSPACE_RGB, True, 8, 
            img.get_width(), img.get_height(), 
            img.get_stride())
        arrows_model.append([pixbuf, k])
        # Check if the style is in use
        if config['arrow_style'] == k:
            active=i
        i += 1
    combo = gtk.ComboBox(arrows_model)
    cell_img = gtk.CellRendererPixbuf()
    combo.pack_start(cell_img)
    combo.add_attribute(cell_img, 'pixbuf', 0)
    cell_text = gtk.CellRendererText()
    cell_text.set_property('xalign', 0.0)
    cell_text.set_property('size-points', 8)
    cell_text.set_property('foreground', '#444444')
    combo.pack_start(cell_text)
    combo.add_attribute(cell_text, 'text', 1)
    print 'Setting combo active', active
    combo.set_active(active)
    controls.pack_start(Label('Arrows'), False)
    controls.pack_start(combo, False, False)
    def changed1(combo):
        idx = combo.get_active()
        config['arrow_style'] = arrow_styles.keys()[idx-1]
        if config['painting']:
            update_buffer()
    combo.connect('changed', changed1)
    # Distributed velocity option
    if not options.of_mode:
        add_check('Multi', 'multi')
    add_check('OF bg.', 'of_bg')
    add_check('C.map', 'color_disk')
    # Color code
    combo = combo_box_new_text()
    for i, cmap in enumerate(options.cmaps):
        combo.append_text(cmap)
        if options.cmaps[cmap] == options.cmap:
            active = i
    combo.set_active(active)
    controls.pack_start(combo, False, False)
    def changed2(combo):
        options.cmap = options.cmaps[combo.get_active_text()]
        if config['painting']:
            update_buffer()
    combo.connect('changed', changed2)
    # Stop painting button
    if config['painting']:
        stock = gtk.STOCK_MEDIA_PAUSE
    else:
        stock = gtk.STOCK_MEDIA_PLAY
    button = gtk.Button(stock=stock)
    def clicked(but):
        config['painting'] = not config['painting']
        if config['painting']:
            stock = gtk.STOCK_MEDIA_PAUSE
            but.set_label(stock)
            update_buffer()
        else:
            stock = gtk.STOCK_MEDIA_PLAY
            but.set_label(stock)
    button.connect('clicked', clicked)
    controls.pack_start(button, False, False)

    # Left toolbox
    paned = gtk.HPaned()
    lbox = VBox()
    paned.pack1(lbox)
    vbox.pack_start(paned, True)

    # Vspace widgets
    if options.frame_vspace or options.pos_vspace:
        lbox.pack_start(gtk.Label('Interp.'), False)
        combo = combo_box_new_text()
        for i in ['linear', 'cubic', 'quintic', 'grid']:
            combo.append_text(i)
        combo.set_active(0)
        lbox.pack_start(combo, False, False)
        def changed(combo):
            if options.frame_vspace:
                f_vspace.set_interpolation(combo.get_active_text())
            if options.pos_vspace:
                p_vspace.set_interpolation(combo.get_active_text())
        combo.connect('changed', changed)
    if options.frame_vspace:
        lbox.pack_start(gtk.Label('Frame'), False)
        t = int(time_scale.get_value())
        f_vspace = VSpaceViewer(options.vspace, dset[t].sum(1).sum(1))
        lbox.pack_start(f_vspace)
    if options.pos_vspace:
        lbox.pack_start(gtk.Label('Pos.'), False)
        p_vspace = VSpaceViewer(options.vspace)
        lbox.pack_start(p_vspace)
    # DrawingArea displaying the visu
    scroll = ScrolledWindow()
    paned.pack2(scroll)
    area = DrawingArea()
    scroll.add_with_viewport(area)
    def expose_cb(area, evt):
        cr = area.window.cairo_create()
        cr.set_source_surface(options.sur)
        cr.paint()
    area.connect('expose_event', expose_cb)

    # Dynamic display
    status = gtk.Statusbar()
    status.pos_msg = status.get_context_id('Position')
    vbox.pack_start(status, False)
    area.add_events(gtk.gdk.POINTER_MOTION_MASK)
    def update_position(area, evt):
        x = max(0, min(width_in - 1, int(evt.x/options.scaling)))
        y = max(0, min(height_in - 1, int(evt.y/options.scaling)))
        t = int(time_scale.get_value())

        status.pop(status.pos_msg)
        status.push(status.pos_msg, '(%d,%d)' % (x, y))

        # Update the velocity space
        if options.pos_vspace:
            p_vspace.set_data(dset[t,:,y,x])

    area.connect("motion_notify_event", update_position)
    
    # Buffering strategy
    width_out, height_out = [int(options.scaling * x) \
                             for x in [width_in, height_in]]
    area.set_size_request(width_out, height_out)
    options.sur = sur = ImageSurface(FORMAT_ARGB32, width_out, height_out)
    options.cr = Context(sur)
    def update_buffer():
        t = int(time_scale.get_value())
        print 'Frame', t
        if options.of_bg or not options.multi:
            print '  OF...'
            if options.of_mode:
                options.of = dset[t,0], dset[t,1]
            else:
                options.of = calculate_of(dset[t], options.vspace,
                                          float(options.of_threshold))
        else:
            options.of = None
        print '  Scaling:', options.scaling
        paint_flow(options.cr, dset[t], options.of, options.vspace,
                   multi=options.multi, of_bg=options.of_bg,
                   color_disk=options.color_disk,
                   scaling=options.scaling, sampling=options.sampling,
                   arrow_length=options.arrow_length,
                   threshold=float(options.activity_threshold),
                   vthreshold=float(options.velocity_threshold),
                   config=config)
        area.queue_draw()
    
    # Start the visualization
    print 'Visu'
    update_buffer()
    # Give a size to the window
    win.set_default_size(gtk.gdk.screen_width() * 3 / 4,
                         gtk.gdk.screen_height() * 3 / 4)
    win.show_all()
    main()

if __name__ == '__main__':
    # Manage CLI arguments
    parser = OptionParser()
    parser.add_option('', '--no-scipy', default=True, dest='scipy',
                      action='store_false',
                      help='Do not use SciPy')
    parser.add_option('-F', '--no-frame-vspace', default=True, dest='frame_vspace',
                      action='store_false',
                      help='Do not display a vspace for each frame')
    parser.add_option('-P', '--no-position-vspace', default=True, dest='pos_vspace',
                      action='store_false',
                      help='Do not display a vspace for each position')
    parser.add_option('-a', '--arrow-style', default='triangles',
                      dest='arrow_style',
                      help='Specify an arrow style')
    parser.add_option('-b', '--begin', default=0, dest='begin',
                      help='First frame to be painted')
    parser.add_option('-c', '--color-disk', default=False, dest='color_disk',
                      action='store_true',
                      help='Draw colormap disk')
    parser.add_option('-d', '--dataset', default=None, dest='dataset',
                      help='Input HDF5 source dataset')
    parser.add_option('-e', '--end', default=None, dest='end',
                      help='First frame not to be painted')
    parser.add_option('-g', '--no-gui', default=True, dest='gui',
                      action='store_false',
                      help='Disable the interactive GUI')
    parser.add_option('-i', '--input', default=None, dest='input',
                      help='Input HDF5 file')
    parser.add_option('-l', '--arrow-length', default=1, dest='arrow_length',
                      help='Arrow length in percetange of output width')
    parser.add_option('-m', '--color-map', default='middleburry', dest='cmap',
                      help='Color map for the OF (middleburry, hls)')
    parser.add_option('-M', '--multi', default=False, dest='multi',
                      action='store_true',
                      help='Display multiple velocities')
    parser.add_option('-o', '--output', default='of-%04d.png', dest='output',
                      help='Output filenames format')
    parser.add_option('-r', '--of-threshold', default=None,
                      dest='of_threshold',
                      help='Activity threshold for the OF (default to vdim*threshold)')
    parser.add_option('-s', '--sampling', default=4, dest='sampling',
                      help='Spatial sampling in input dataset')
    parser.add_option('-t', '--threshold', default=0.001,
                      dest='activity_threshold',
                      help='Activity threshold')
    parser.add_option('-v', '--velocity-threshold', default=0.001,
                      dest='velocity_threshold',
                      help='Velocity norm threshold')
    parser.add_option('-w', '--white-background', default=True, dest='of_bg',
                      action='store_false',
                      help='Use white background instead of colored OF')
    parser.add_option('-z', '--scaling', default=2.0, dest='scaling',
                      help='Input to output scaling factor for each dim.')
    parser.add_option('', '--max-veloc', default=None, dest='maxvel',
                      help='Maximum acceptable velocity')
    options, args = parser.parse_args()
    
    # Check CLI arguments
    if not options.input:
        if len(args) >= 1:
            options.input = args[0]
        else:
            exit('No input HDF5 file specified')
    if not options.dataset:
        if len(args) >= 2:
            options.dataset = args[1]
        else:
            if options.gui:
                # Ask which dataset to open
                import h5py
                hf = h5py.File(options.input)
                dsets = hf.keys()
                print 'Not dataset specified, please enter one:'
                print ' '.join(dsets)
                options.dataset = raw_input('Dataset: ')
                del hf
            else:
                exit('Not input HDF5 dataset specified')
    options.cmaps = {'middleburry' : middleburry,
                     'hls' : hls}
    if not options.cmap in options.cmaps:
        options.cmap = 'middleburry'
    options.cmap = options.cmaps[options.cmap]

    # Check for SciPy
    if options.scipy and (options.pos_vspace or options.frame_vspace):
        try:
            import scipy
        except ImportError:
            print >>stderr, '*** Could NOT import SciPy, disabling vspace visualisation ***'
            options.pos_vspace = False
            options.frame_vspace = False
    else:
        options.pos_vspace = False
        options.frame_vspace = False

    # Create a configuration
    config = {
        'painting': True,
    }
    if options.arrow_style in arrow_styles:
        config['arrow_style'] = options.arrow_style
    else:
        config['arrow_style'] = None
        if options.arrow_style != 'none':
            print >>sys.stderr, 'Unknown arrow style: %s' % options.arrow_style
    
    # Open an HDF reference to the dataset
    hf = openFile(options.input, 'r')
    if not hasattr(hf.root, options.dataset):
        exit('No such dataset: %s' % options.dataset)
    dset = getattr(hf.root, options.dataset)
    if len(dset.shape) == 4:
        depth, vdim, height_in, width_in = dset.shape
    else:
        exit('Unmanaged dataset dimensions: %s' % str(dset.shape))
    
    # Check again CLI arguments
    if options.end is None:
        options.end = depth
    if options.of_threshold is None:
        options.of_threshold = 0.001 #vdim * float(options.activity_threshold)
        
    # Create the output image surface
    options.scaling = float(options.scaling)
    width_out, height_out = [int(options.scaling * x) \
                             for x in [width_in, height_in]]
    options.arrow_length = float(options.arrow_length) #/ 100. * width_out
    options.sampling = int(options.sampling)
    if vdim != 2:
        # Check for stored vspace
        if hasattr(dset._v_attrs, 'vspace'):
            options.vspace = dset._v_attrs.vspace
        else:
            options.vspace = parse_vspace(int(sqrt(vdim) / 2))

    # Optical-flow only mode
    if vdim == 2:
        options.multi = False
        options.triangles = True
        options.of_mode = True
        options.vspace = []
        options.frame_vspace = False
        options.pos_vspace = False
    # Normal full vspace mode
    else:
        options.of_mode = False
    
    # GUI mode
    if options.gui:
        run_gui(dset, options, config)
    # Non-interactive mode
    else:
        sur = ImageSurface(FORMAT_ARGB32, width_out, height_out)
        cr = Context(sur)
        for t in xrange(int(options.begin), int(options.end)):
            print 'Painting frame %d' % t
            print dset[t].shape
            if options.of_mode:
                of = dset[t]
            else:
                if options.of_bg or not options.multi:
                    print '  OF...'
                    of = calculate_of(dset[t], options.vspace, float(options.of_threshold))
                else:
                    of = None
            # Cut values too high
            if options.maxvel:
                mask = (abs(of[0]) < float(options.maxvel)) * (abs(of[1]) < float(options.maxvel))
                of[0] *= mask
                of[1] *= mask
            paint_flow(cr, dset[t], of, options.vspace,
                       multi=options.multi, of_bg=options.of_bg,
                       color_disk=options.color_disk,
                       scaling=options.scaling, sampling=options.sampling,
                       arrow_length=options.arrow_length,
                       threshold=float(options.activity_threshold),
                       vthreshold=float(options.velocity_threshold),
                       config=config)
            print '  Writing...'
            sur.write_to_png(options.output % t)
            print
